"""TROC Backend.

Troc Authentication using RNC algorithm.
"""
import jwt
from aiohttp import web
import rapidjson
from .base import BaseAuthBackend
from navigator.libs.cypher import *
from datetime import datetime, timedelta
from aiohttp_session import get_session, new_session
from navigator.exceptions import NavException, UserDoesntExists, InvalidAuth
from navigator.conf import (
    PARTNER_KEY,
    CYPHER_TYPE,
    SESSION_TIMEOUT,
    SECRET_KEY
)
import hashlib
import base64
import secrets

# TODO: add expiration logic when read the token
CIPHER = Cipher(PARTNER_KEY, type=CYPHER_TYPE)


class TrocAuth(BaseAuthBackend):
    """TROC authentication Header."""
    user_attribute: str = 'user'
    username_attribute: str = 'email'
    _scheme: str = 'Bearer'

    def __init__(
            self,
            user_property: str = 'user',
            user_attribute: str = 'user',
            userid_attribute: str = 'user_id',
            username_attribute: str = 'email',
            credentials_required: bool = False,
            authorization_backends: tuple = (),
            session_type: str = 'cookie',
            **kwargs
    ):
        super().__init__(
            user_property,
            user_attribute,
            userid_attribute,
            username_attribute,
            credentials_required,
            authorization_backends,
            session_type,
            **kwargs
        )
        # forcing to use Email as Username Attribute
        self.username_attribute = 'email'

    async def validate_user(self, login: str = None):
        # get the user based on Model
        search = {
            self.username_attribute: login
        }
        try:
            user = await self.get_user(**search)
            return user
        except UserDoesntExists as err:
            raise UserDoesntExists(f'User {login} doesnt exists')
        except Exception as err:
            raise Exception(err)
        return None

    async def get_payload(self, request):
        troctoken = None
        try:
            if 'Authorization' in request.headers:
                try:
                    scheme, troctoken = request.headers.get(
                        'Authorization'
                    ).strip().split(' ')
                except ValueError:
                    raise web.HTTPForbidden(
                        reason='Invalid authorization Header',
                    )
                if scheme != self._scheme:
                    raise web.HTTPForbidden(
                        reason='Invalid Session scheme',
                    )
            else:
                try:
                    troctoken = request.query.get('auth', None)
                except Exception as e:
                    print(e)
                    return None
        except Exception as e:
            print(e)
            return None
        return troctoken

    async def check_credentials(self, request):
        try:
            troctoken = await self.get_payload(request)
        except Exception as err:
            raise NavException(err, state=400)
        if not troctoken:
            raise InvalidAuth('Invalid Credentials', state=401)
        else:
            # getting user information
            # TODO: making the validation of token and expiration
            try:
                data = rapidjson.loads(CIPHER.decode(passphrase=troctoken))
            except Exception as err:
                raise InvalidAuth(f'Invalid TROC Token: {err!s}', state=401)
            # making validation
            try:
                username = data[self.username_attribute]
            except KeyError as err:
                raise InvalidAuth(f'Missing Email attribute: {err!s}', state=401)
            try:
                user = await self.validate_user(
                    login=username
                )
            except UserDoesntExists as err:
                raise UserDoesntExists(err)
            except Exception as err:
                raise NavException(err, state=500)
            try:
                userdata = self.get_userdata(user)
                userdata['Payload'] = data
                # Create the User session and returned.
                session = await self._session.create_session(
                    request,
                    user,
                    userdata
                )
                payload = {
                    self.user_property: user[self.userid_attribute],
                    self.username_attribute: user[self.username_attribute],
                    'user_id': user[self.userid_attribute]
                }
                token = self.create_jwt(data=payload)
                return {'token': token}
            except Exception as err:
                print(err)
                return False

    async def authenticate(self, request):
        """ Authenticate, refresh or return the user credentials."""
        pass

    async def auth_middleware(self, app, handler):
        async def middleware(request):
            request.user = None
            authz = await self.authorization_backends(app, handler, request)
            if authz:
                return await authz
            try:
                jwt_token = self.decode_token(request)
            except NavException as err:
                response = {
                    "message": "TROC Token Error",
                    "error": err.message,
                    "status": err.state
                }
                return web.json_response(response, status=err.state)
            except Exception as err:
                raise web.HTTPBadRequest(
                    body=f'Bad Request: {err!s}'
                )
            if self.credentials_required is True:
                raise web.HTTPUnauthorized(
                    body='Unauthorized'
                )
            return await handler(request)
        return middleware



"""Django Session Backend.

Navigator Authentication using Django Session Backend
"""
import logging
import asyncio
from aiohttp import web, hdrs
from .base import BaseAuthBackend
import uuid

class NoAuth(BaseAuthBackend):
    """Basic Handler for No authentication."""
    user_attribute: str = 'userid'

    async def check_credentials(self, request):
        """ Authentication and create a session."""
        return True

    async def authenticate(self, request):
        return True

    async def get_session(self, request):
        return {}

    async def auth_middleware(self, app, handler):
        async def middleware(request):
            return await handler(request)
        return middleware



"""TROC Token Middleware.

Use RNC algorithm to create a token-based authentication/authorization
for Navigator.

Middleware Authorization.
"""
import sys
import json
from aiohttp import web
from navigator.libs.cypher import *

from navigator.conf import (
    PARTNER_KEY,
    CYPHER_TYPE,
    PARTNER_SESSION_TIMEOUT
)

# TODO: add expiration logic when read the token
CIPHER = Cipher(PARTNER_KEY, type=CYPHER_TYPE)


async def troctoken_middleware(app, handler):
    async def middleware(request):
        request.user = None
        try:
            troctoken = request.query.get('auth', request.headers.get('X-Token', None))
        except KeyError as err:
            troctoken = None
        if troctoken:
            try:
                payload = CIPHER.decode(passphrase=troctoken)
            except ValueError as err:
                print(err)
                raise web.HTTPUnauthorized(
                    reason='Decryption authorization Error'
                )
            except Exception as err:
                print(err)
                return web.json_response(
                    {'message': f"Token Error {err!s}"}, status=500)
            if not payload:
                raise web.HTTPForbidden(reason='Invalid authorization Token')
            # TODO: make the validation of the token
        return await handler(request)
    return middleware



"""Memcached-based Session Object."""
import asyncio
import rapidjson
from functools import wraps, partial
# memcached
import aiomcache
from aiohttp_session.memcached_storage import MemcachedStorage
from aiohttp_session import setup as setup_session
from .base import AbstractSession
from navigator.conf import (
    DOMAIN,
    SESSION_URL,
    SESSION_TIMEOUT,
    MEMCACHE_HOST,
    MEMCACHE_PORT
)


class MemcacheSession(AbstractSession):
    """Session Storage based on Memcache."""
    _pool = None

    async def get_mcache(self, **kwargs):
        loop = asyncio.get_event_loop()
        return aiomcache.Client(MEMCACHE_HOST, MEMCACHE_PORT, loop=loop)

    def configure_session(self, app, **kwargs):
        async def _make_mcache():
            _encoder = partial(rapidjson.dumps, datetime_mode=rapidjson.DM_ISO8601)
            try:
                self._pool = await self.get_mcache()
                setup_session(
                    app,
                    MemcachedStorage(
                        self._pool,
                        cookie_name=self.session_name,
                        encoder=_encoder,
                        decoder=rapidjson.loads,
                        max_age=int(SESSION_TIMEOUT)
                    )
                )
            except Exception as err:
                print(err)
                return False
        return asyncio.get_event_loop().run_until_complete(
            _make_mcache()
        )



import base64
import rapidjson
from aiohttp import web
from datetime import datetime, timedelta
from navigator.conf import (
    SESSION_TIMEOUT,
    SECRET_KEY,
    SESSION_PREFIX
)

async def django_middleware(app, handler):
    async def middleware(request):
        request.user = None
        try:
            session = request.headers.get("X-Sessionid", None)
        except Exception as e:
            print(e)
            session = request.headers.get("Sessionid", None)
        if session:
            redis = app['redis']
            try:
                result = await redis.get("{}:{}".format(SESSION_PREFIX, session))
                if not result:
                    return web.json_response(
                        {'message': 'Invalid Django Session'}, status=400
                    )
                try:
                    data = base64.b64decode(result)
                    session_data = data.decode("utf-8").split(":", 1)
                    user = rapidjson.loads(session_data[1])
                    session = {
                        "key": session,
                        "session_id": session_data[0],
                        **user
                    }
                    request['user_id'] = user['user_id']
                    request['session'] = session
                except Exception as err:
                    print(err)
                    return web.json_response(
                        {'message': 'Error Decoding Django Session'}, status=400
                    )
            except Exception as err:
                print(err)
                return web.json_response(
                    {'message': 'Invalid Session'}, status=400
                )
        return await handler(request)
    return middleware



"""Session Object for Navigator."""

from .base import AbstractSession
from .cookie import CookieSession
from .redis import RedisSession
from .memcache import MemcacheSession


__all__ = ["AbstractSession", "CookieSession", "RedisSession", "MemcacheSession"]



"""Navigator Auth.

Navigator Authentication/Authorization system.
"""
from textwrap import dedent
import importlib
import logging
from aiohttp import web
from typing import List, Iterable
from .backends import BaseAuthBackend
# aiohttp session
from .authorizations import *
from navigator.functions import json_response
from navigator.exceptions import NavException, UserDoesntExists, InvalidAuth

class AuthHandler(object):
    """Authentication Backend for Navigator."""
    _template = """
        <!doctype html>
            <head></head>
            <body>
                <p>{message}</p>
                <form action="/login" method="POST">
                  Login:
                  <input type="text" name="login">
                  Password:
                  <input type="password" name="password">
                  <input type="submit" value="Login">
                </form>
                <a href="/logout">Logout</a>
            </body>
    """
    backend = None
    _session = None
    _user_property: str = 'user'
    _required: bool = False

    def __init__(
            self,
            backend: str = 'navigator.auth.backends.NoAuth',
            credentials_required: bool = False,
            auth_scheme='Bearer',
            authorization_backends: List = (),
            **kwargs
    ):
        self._template = dedent(self._template)
        authz_backends = self.get_authorization_backends(
            authorization_backends
        )
        args = {
            "credentials_required": credentials_required,
            "scheme": auth_scheme,
            "authorization_backends": authz_backends,
            **kwargs
        }
        self.backend = self.get_backend(backend, **args)


    def get_backend(self, backend, **kwargs):
        try:
            parts = backend.split('.')
            bkname = parts[-1]
            classpath = '.'.join(parts[:-1])
            module = importlib.import_module(classpath, package=bkname)
            obj = getattr(module, bkname)
            return obj(**kwargs)
        except ImportError:
            raise Exception(f"Error loading Auth Backend {backend}")

    async def close(self):
        """
        Cleanup the processes
        """
        pass

    def get_authorization_backends(self, backends: Iterable) -> tuple:
        b = []
        for backend in backends:
            # TODO: more automagic logic
            if backend == 'hosts':
                b.append(authz_hosts())
            elif backend == 'allow_hosts':
                b.append(authz_allow_hosts())
        return b

    # async def login(self, request) -> web.Response:
    #     response = web.HTTPFound("/")
    #     form = await request.post()
    #     login = form.get("login")
    #     password = form.get("password")
    #     if user := await self.check_credentials(login, password):
    #         # if state, save user data in session
    #         state = await self._session.create_session(request, user=user)
    #         raise response
    #     else:
    #         template = self._template.format(
    #             message="Invalid =username/password= combination"
    #         )
    #     raise web.HTTPUnauthorized(text=template, content_type="text/html")
    #
    # async def login_page(self, request):
    #     username = None
    #     # check if authorized, instead, return to login
    #     #session = await get_session(request)
    #     # try:
    #     #     username = session["username"]
    #     # except KeyError:
    #     #     template = self._template.format(message="You need to login")
    #     # print(template)
    #     # if username:
    #     #     template = self._template.format(
    #     #         message="Hello, {username}!".format(username=username)
    #     #     )
    #     # else:
    #     #     template = self._template.format(message="You need to login")
    #     # print(template)
    #     template = self._template.format(message="You need to login")
    #     return web.Response(text=template, content_type="text/html")
    #
    # async def logout(self, request: web.Request) -> web.Response:
    #     await self._session.forgot_session(request)
    #     raise web.HTTPSeeOther(location="/")

    async def api_logout(self, request: web.Request) -> web.Response:
        await self.backend.forgot_session(request)
        return web.json_response({"message": "logout successful"}, status=200)

    async def api_login(self, request: web.Request) -> web.Response:
        try:
            user = await self.backend.check_credentials(request)
            if not user:
                raise web.HTTPUnauthorized(
                    reason='Unauthorized',
                    status=403
                )
            return json_response(user, state=200)
        except (NavException, UserDoesntExists, InvalidAuth) as err:
            raise web.HTTPUnauthorized(
                reason=err,
                status=err.state
            )
        except ValueError:
            raise web.HTTPUnauthorized(
                reason='Unauthorized'
            )
        except Exception as err:
            print(err)
            raise web.HTTPUnauthorized(reason=err, status=403)

    async def authenticate(self, request: web.Request) -> web.Response:
        """ Authentication method to refresh credentials for Registration."""
        auth = await self.backend.check_authorization(request)
        if not auth:
            raise web.HTTPUnauthorized(
                reason='User not Authorized'
            )

    async def get_session(self, request: web.Request) -> web.Response:
        """ return Session Data from user."""
        try:
            dump = await self.backend.get_session(request)
        except NavException as err:
            print('Error HERE: ', err, err.state)
            response = {
                "message": "Session Error",
                "error": err.message,
                "status": err.state
            }
            return web.json_response(response, status=err.state)
        if dump:
            return json_response(dump)
        else:
            raise web.HTTPForbidden()

    def configure(self, app: web.Application) -> web.Application:
        router = app.router
        # router.add_route("GET", "/login", self.login_page, name="index_login")
        # router.add_route("POST", "/login", self.login, name="login")
        # router.add_route("GET", "/logout", self.logout, name="logout")
        router.add_route("GET", "/api/v1/login/{program}", self.api_login, name="api_login_get_tenant")
        router.add_route("POST", "/api/v1/login/{program}", self.api_login, name="api_login_post_tenant")
        router.add_route("GET", "/api/v1/login", self.api_login, name="api_login_get")
        router.add_route("POST", "/api/v1/login", self.api_login, name="api_login_post")
        router.add_route("GET", "/api/v1/logout", self.api_logout, name="api_logout")
        router.add_route("GET", "/api/v1/authenticate/{program}", self.authenticate, name="api_authenticate_program")
        router.add_route("GET", "/api/v1/authenticate", self.authenticate, name="api_authenticate")
        router.add_route("GET", "/api/v1/session/{program}", self.get_session, name="api_session_tenant")
        router.add_route("GET", "/api/v1/session", self.get_session, name="api_session")
        # backed needs initialization (connection to a redis server, etc)
        try:
            self.backend.configure(app, router)
        except Exception as err:
            print(err)
            logging.exception(f'Error on Auth Backend initialization {err!s}')
        # the backend add a middleware to the app
        mdl = app.middlewares
        mdl.append(self.backend.auth_middleware)
        return app



"""Django Session Backend.

Navigator Authentication using Django Session Backend
description: read the Django session from Redis Backend and decrypt.
"""
import base64
import rapidjson
import logging
import asyncio
import jwt
from aiohttp import web, hdrs
from asyncdb import AsyncDB
from .base import BaseAuthBackend
from navigator.exceptions import NavException, UserDoesntExists, InvalidAuth
from datetime import datetime, timedelta
from navigator.conf import (
    SESSION_URL,
    SESSION_TIMEOUT,
    SECRET_KEY,
    PARTNER_KEY,
    JWT_ALGORITHM,
    SESSION_PREFIX,
    default_dsn
)


class TokenAuth(BaseAuthBackend):
    """API Token Authentication Handler."""
    connection = None
    _scheme: str = 'Bearer'

    def configure(self, app, router):
        async def _make_connection():
            try:
                self.connection = AsyncDB('pg', dsn=default_dsn)
                await self.connection.connection()
            except Exception as err:
                print(err)
                raise Exception(err)
        asyncio.get_event_loop().run_until_complete(
            _make_connection()
        )
        # executing parent configurations
        super(TokenAuth, self).configure(app, router)

    async def get_payload(self, request):
        token = None
        tenant = None
        id = None
        try:
            if 'Authorization' in request.headers:
                try:
                    scheme, id = request.headers.get(
                        'Authorization'
                    ).strip().split(' ', 1)
                except ValueError:
                    raise web.HTTPForbidden(
                        reason='Invalid authorization Header',
                    )
                if scheme != self._scheme:
                    raise web.HTTPForbidden(
                        reason='Invalid Authorization Scheme',
                    )
                try:
                    tenant, token = id.split(':')
                except ValueError:
                    raise web.HTTPForbidden(
                        reason='Invalid Token Structure',
                    )
        except Exception as e:
            print(e)
            return None
        return [tenant, token]

    async def reconnect(self):
        if not self.connection or not self.connection.is_connected():
            await self.connection.connection()

    async def check_credentials(self, request):
        try:
            tenant, token = await self.get_payload(request)
            logging.debug(f'Tenant ID: {tenant}')
        except Exception as err:
            raise NavException(err, state=400)
        if not token:
            raise InvalidAuth('Invalid Credentials', state=401)
        else:
            payload = jwt.decode(
                token,
                PARTNER_KEY,
                algorithms=[JWT_ALGORITHM],
                leeway=30
            )
            logging.debug(f'Decoded Token: {payload!s}')
            data = await self.check_token_info(request, tenant, payload)
            if not data:
                raise InvalidAuth(f'Invalid Session: {token!s}', state=401)
            # getting user information
            # making validation
            try:
                u = data['name']
                username = data['partner']
                grants = data['grants']
                programs = data['programs']
            except KeyError as err:
                print(err)
                raise InvalidAuth(
                    f'Missing attributes for Partner Token: {err!s}',
                    state=401
                )
            # TODO: Validate that partner (tenants table):
            # try:
            #     user = await self.validate_user(
            #         login=username
            #     )
            # except UserDoesntExists as err:
            #     raise UserDoesntExists(err)
            # except Exception as err:
            #     raise NavException(err, state=500)
            try:
                user = {
                  'name': data['name'],
                  'partner': username,
                  'issuer': 'Mobileinsight',
                  'programs': programs,
                  'grants': grants,
                  'tenant': tenant
                }
                userdata = dict(data)
                # Create the User session and returned.
                session = await self._session.create_session(
                    request,
                    user,
                    userdata
                )
                session['partner'] = username
                session['tenant'] = tenant
                session['programs'] = programs
                token = self.create_jwt(data=user)
                return {'token': f'{tenant}:{token}'}
            except Exception as err:
                print(err)
                return False

    async def authenticate(self, request):
        """ Authenticate, refresh or return the user credentials."""
        pass

    async def check_token_info(self, request, tenant, payload):
        if not self.connection:
            await self.reconnect()
        try:
            name = payload['name']
            partner = payload['partner']
        except KeyError as err:
            return False
        sql = f"""
        SELECT name, partner, grants, programs FROM troc.api_keys
        WHERE name='{name}' AND partner='{partner}'
        AND enabled = TRUE AND revoked = FALSE AND '{tenant}'= ANY(programs)
        """
        try:
            result, error = await self.connection.queryrow(sql)
            if error or not result:
                return False
            else:
                return result
        except Exception as err:
            logging.exception(err)
            return False

    async def auth_middleware(self, app, handler):
        async def middleware(request):
            request.user = None
            authz = await self.authorization_backends(app, handler, request)
            if authz:
                return await authz
            tenant, token = await self.get_payload(request)
            if token:
                try:
                    payload = jwt.decode(
                        token,
                        PARTNER_KEY,
                        algorithms=[JWT_ALGORITHM],
                        leeway=30
                    )
                    logging.debug(f'Decoded Token: {payload!s}')
                    result = await self.check_token_info(request, tenant, payload)
                    if not result:
                        raise web.HTTPForbidden(
                            reason='Not Authorized',
                        )
                    else:
                        session = await self._session.get_session(request)
                        session['grants'] = result['grants']
                        session['partner'] = result['partner']
                        session['tenant'] = tenant
                except (jwt.DecodeError) as err:
                    raise web.HTTPBadRequest(
                        reason=f'Token Decoding Error: {err!r}'
                    )
                except jwt.InvalidTokenError as err:
                    print(err)
                    raise web.HTTPBadRequest(
                        reason=f'Invalid authorization token {err!s}'
                    )
                except (jwt.ExpiredSignatureError) as err:
                    print(err)
                    raise web.HTTPBadRequest(
                        reason=f'Token Expired: {err!s}'
                    )
                except Exception as err:
                    print(err, err.__class__.__name__)
                    raise web.HTTPBadRequest(
                        reason=f'Bad Authorization Request: {err!s}'
                    )
            else:
                if self.credentials_required is True:
                    print('Missing Token information')
                    raise web.HTTPUnauthorized(
                        reason='Not Authorized',
                    )
            return await handler(request)
        return middleware



"""
Model System for Navigator Auth.

Model for User, Group and Roles for Navigator Auth.
"""
import uuid
import asyncio
import logging
from asyncdb.utils.models import Model, Column
from typing import Optional, List, Dict, Union, Tuple, Any, Callable
from dataclasses import InitVar
from datetime import datetime
from navigator.conf import default_dsn, USERS_TABLE

class User(Model):
    """Basic User notation."""
    user_id: int = Column(required=False, primary_key=True)
    first_name: str
    last_name: str
    email: str = Column(required=False, max=254)
    password: str = Column(required=False, max=128)
    last_login: datetime = Column(required=False)
    username: str = Column(required=False)
    is_superuser: bool = Column(required=True, default=False)
    is_active: bool = Column(required=True, default=True)
    is_new: bool = Column(required=True, default=True)
    title: str = Column(equired=False, max=90)
    registration_key: str = Column(equired=False, max=512)
    reset_pwd_key: str = Column(equired=False, max=512)
    avatar: str = Column(max=512)

    def __getitem__(self, item):
        return getattr(self, item)

    @property
    def display_name(self):
        return f"{self.first_name} {self.last_name}"

    class Meta:
        driver = 'pg'
        dsn: str = default_dsn
        name = USERS_TABLE
        schema = 'public'
        strict = True
        frozen = False
        connection = None

# TODO: add autoincrement feature, read the last id and plus 1
class Group(Model):
    id: int = Column(required=True, primary_key=True)
    name: int = Column(required=True)

    class Meta:
        driver = 'pg'
        dsn: str = default_dsn
        name = 'auth_group'
        schema = 'public'
        strict = True
        frozen = False
        connection = None

# class UserGroup(Model):
#     member_id: int = Column(required=True, primary_key=True)
#     user_id: User = Column(required=True)
#     group_id: Group = Column(required=True)
#
#     class Meta:
#         driver = 'pg'
#         name = 'auth_membership'
#         schema = 'public'
#         app_label = 'troc'
#         strict = True
#         frozen = False
#
# class Role(Model):
#     role_id: int = Column(required=True, primary_key=True)
#     rolename: str = Column(required=True, unique=True)
#     role: str = Column(required=True, unique=True, comment="Role Slug")
#     program_id: int = Column(required=False, default=1)
#
#     class Meta:
#         driver = 'pg'
#         name = 'auth_roles'
#         schema = 'public'
#         app_label = 'troc'
#         strict = True
#         frozen = False



"""Connection Manager for Navigator."""

import logging
import json
import uuid
from asyncpg.pgproto import pgproto
from asyncdb.utils.encoders import BaseEncoder
from asyncdb import AsyncPool
from asyncdb.providers import BasePool, BaseProvider
from typing import Dict, List, Callable, Optional, Iterable

try:
    from settings.settings import (
        TIMEZONE
    )
except ImportError:
    # Timezone (For parsedate)
    TIMEZONE = 'America/New_York'


class AbstractConnection(object):
    driver: str = 'pg'
    pool_based: bool = True
    _loop = None
    timeout: int = 60
    _init: Callable = None
    _startup: Callable = None
    _shutdown: Callable = None

    def __init__(
        self,
        driver: str = 'pg',
        dsn: str = '',
        name: str = '',
        init: Callable = None,
        startup: Callable = None,
        shutdown: Callable = None,
        **kwargs
    ):
        self.driver = driver
        if 'loop' in kwargs:
            self._loop = kwargs['loop']
            del kwargs['loop']
        if init:
            self._init = init
        if startup:
            self._startup = startup
        if shutdown:
            self._shutdown = shutdown

        if self.pool_based:
            self.conn = AsyncPool(
                self.driver,
                dsn=dsn,
                timeout=self.timeout,
                **kwargs
            )
        else:
            self.conn = AsyncDB(
                self.driver,
                dsn=dsn,
                timeout=self.timeout,
                **kwargs
            )
        # configure connection
        self.configure()

    def connection(self):
        return self.conn

    def configure(self):
        if self._init and self.pool_based:
            self.conn.setup_func = self._init

    def is_connected(self):
        return bool(self._connected)

    async def startup(self, **kwargs):
        if 'app' in kwargs:
            app = kwargs['app']
            if app is not None:
                if 'database' in app:
                    # re-use the database connection
                    self.conn = app['database']
        if self.pool_based:
            await self.conn.connect()
        else:
            await self.conn.connection()
        if self._startup:
            await self._startup(self.conn, **kwargs)
        self._connected = True

    async def shutdown(self, **kwargs):
        if self._shutdown:
            await self._shutdown(self.conn, **kwargs)
        logging.debug('Closing all connections')
        if self.pool_based:
            logging.debug('Closing DB Pool')
            await self.conn.wait_close(gracefully=False, timeout=5)
        else:
            await self.conn.close()
        logging.debug('Exiting ...')


class PostgresPool(AbstractConnection):
    driver: str = 'pg'
    pool_based: bool = True
    timeout: int = 3600

    def __init__(
        self,
        dsn: str,
        name: str = '',
        init: Callable = None,
        startup: Callable = None,
        shutdown: Callable = None,
        **kwargs
    ):
        kwargs = {
            'min_size': 10,
            "server_settings": {
                'application_name': name,
                'client_min_messages': 'notice',
                'max_parallel_workers': '24',
                'jit': 'on',
                'statement_timeout': '3600000',
                'timezone': TIMEZONE
            }
        }
        if 'loop' in kwargs:
            self._loop = kwargs['loop']
            del kwargs['loop']
        if init:
            self._init = init
        if startup:
            self._startup = startup
        if shutdown:
            self._shutdown = shutdown

        self.conn = AsyncPool(
            self.driver,
            dsn=dsn,
            timeout=self.timeout,
            **kwargs
        )
        # passing the configuration
        self.conn.setup_func = self.configure

    async def configure(self, conn):
        # also, if exists this init connection, run
        if self._init:
            await self._init(conn)



"""Django Session Backend.

Navigator Authentication using Django Session Backend
description: read the Django session from Redis Backend and decrypt.
"""
import base64
import rapidjson
import logging
import asyncio
# redis pool
import aioredis
from aiohttp import web, hdrs
from .base import BaseAuthBackend
from navigator.exceptions import NavException, UserDoesntExists, InvalidAuth
from datetime import datetime, timedelta
from navigator.conf import (
    SESSION_URL,
    SESSION_TIMEOUT,
    SECRET_KEY,
    SESSION_PREFIX
)


class SessionIDAuth(BaseAuthBackend):
    """Django SessionID Authentication Handler."""
    redis = None
    _scheme: str = 'Bearer'

    def configure(self, app, router):
        async def _make_redis():
            try:
                self.redis = await aioredis.create_redis_pool(
                    SESSION_URL, timeout=1
                )
            except Exception as err:
                print(err)
                raise Exception(err)
        asyncio.get_event_loop().run_until_complete(
            _make_redis()
        )
        # executing parent configurations
        super(SessionIDAuth, self).configure(app, router)

    async def get_payload(self, request):
        id = None
        try:
            if 'Authorization' in request.headers:
                try:
                    scheme, id = request.headers.get(
                        'Authorization'
                    ).strip().split(' ')
                except ValueError:
                    raise web.HTTPForbidden(
                        reason='Invalid authorization Header',
                    )
                if scheme != self._scheme:
                    raise web.HTTPForbidden(
                        reason='Invalid Session scheme',
                    )
            elif 'X-Sessionid' in request.headers:
                id = request.headers.get('X-Sessionid', None)
        except Exception as e:
            print(e)
            return None
        return id

    async def validate_session(self, key: str = None):
        try:
            result = await self.redis.get("{}:{}".format(SESSION_PREFIX, key))
            if not result:
                return False
            data = base64.b64decode(result)
            session_data = data.decode("utf-8").split(":", 1)
            user = rapidjson.loads(session_data[1])
            session = {
                "key": key,
                "session_id": session_data[0],
                self.user_property: user
            }
            return session
        except Exception as err:
            print(err)
            logging.debug("Django Session Decoding Error: {}".format(err))
            return False

    async def validate_user(self, login: str = None):
        # get the user based on Model
        search = {
            self.userid_attribute: login
        }
        try:
            user = await self.get_user(**search)
            return user
        except UserDoesntExists as err:
            raise UserDoesntExists(f'User {login} doesnt exists')
        except Exception as err:
            raise Exception(err)
        return None

    async def check_credentials(self, request):
        try:
            sessionid = await self.get_payload(request)
            logging.debug(f'Session ID: {sessionid}')
        except Exception as err:
            raise NavException(err, state=400)
        if not sessionid:
            raise InvalidAuth('Invalid Credentials', state=401)
        else:
            # getting user information
            # TODO: making the validation of token and expiration
            try:
                data = await self.validate_session(key=sessionid)
            except Exception as err:
                raise InvalidAuth(f'Invalid Session: {err!s}', state=401)
            # making validation
            try:
                u = data[self.user_property]
                username = u[self.userid_attribute]
            except KeyError as err:
                print(err)
                raise InvalidAuth(
                    f'Missing {self.userid_attribute} attribute: {err!s}',
                    state=401
                )
            try:
                user = await self.validate_user(
                    login=username
                )
            except UserDoesntExists as err:
                raise UserDoesntExists(err)
            except Exception as err:
                raise NavException(err, state=500)
            try:
                userdata = self.get_userdata(user)
                userdata['session'] = data
                # Create the User session and returned.
                session = await self._session.create_session(
                    request,
                    user,
                    userdata
                )
                payload = {
                    self.user_property: user[self.userid_attribute],
                    self.username_attribute: user[self.username_attribute],
                    'user_id': user[self.userid_attribute]
                }
                token = self.create_jwt(data=payload)
                return {'token': token}
            except Exception as err:
                print(err)
                return False

    async def authenticate(self, request):
        """ Authenticate, refresh or return the user credentials."""
        pass

    async def auth_middleware(self, app, handler):
        async def middleware(request):
            request.user = None
            authz = await self.authorization_backends(app, handler, request)
            if authz:
                return await authz
            if 'Authorization' in request.headers:
                try:
                    jwt_token = self.decode_token(request)
                except NavException as err:
                    response = {
                        "message": "Session ID Failure",
                        "error": err.message,
                        "status": err.state
                    }
                    return web.json_response(response, status=err.state)
                except Exception as err:
                    raise web.HTTPBadRequest(
                        body=f'Bad Request: {err!s}'
                    )
                if self.credentials_required is True:
                    raise web.HTTPUnauthorized(
                        body='Unauthorized'
                    )
            elif 'X-Sessionid' in request.headers:
                sessionid = request.headers.get('X-Sessionid', None)
                try:
                    data = await self.validate_session(key=sessionid)
                    # TODO: making validation using only sessionid
                except Exception as err:
                    raise web.HTTPUnauthorized(
                        reason='Invalid Authorization Session',
                    )
            else:
                if self.credentials_required is True:
                    raise web.HTTPUnauthorized(
                        reason='Missing Authorization Session',
                    )
            return await handler(request)
        return middleware



"""Cookie-based Session Storage."""

from aiohttp_session.cookie_storage import EncryptedCookieStorage
from .base import AbstractSession
from aiohttp_session import setup as setup_session
from navigator.conf import (
    DOMAIN,
    SESSION_TIMEOUT
)


class CookieSession(AbstractSession):
    """Encrypted Cookie Session Storage."""

    def configure_session(self, app):
        """Cookie Session Configuration."""
        print(self.secret_key, self.session_name)
        setup_session(
            app,
            EncryptedCookieStorage(
                self.secret_key,
                cookie_name=self.session_name,
                max_age=SESSION_TIMEOUT
                # domain=DOMAIN,
            )
        )



"""JWT Backend.

Navigator Authentication using JSON Web Tokens.
"""
import jwt
from aiohttp import web
from .base import BaseAuthBackend
from datetime import datetime, timedelta
from aiohttp_session import get_session, new_session
from navigator.exceptions import NavException, UserDoesntExists, InvalidAuth
from navigator.conf import (
    SESSION_TIMEOUT,
    SECRET_KEY
)
import hashlib
import base64
import secrets

# credentials algorithm
PWD_ALGORITHM = "pbkdf2_sha256"
ITERATIONS = 150000
PWD_DIGEST = 'sha256'
KEY_LENGTH = 64

# "%s$%d$%s$%s" % (algorithm, iterations, salt, hash)

class BasicAuth(BaseAuthBackend):
    """Basic User/pasword with JWT Authentication."""
    user_attribute: str = 'user'
    username_attribute: str = 'username'
    pwd_atrribute: str = 'password'
    scheme: str = 'Basic'

    async def validate_user(self, login: str = None, password: str = None):
        # get the user based on Model
        search = {
            self.username_attribute: login
        }
        try:
            user = await self.get_user(**search)
        except UserDoesntExists as err:
            raise UserDoesntExists(f'User {login} doesnt exists')
        except Exception as err:
            raise Exception(err)
        try:
            # later, check the password
            pwd = user[self.pwd_atrribute]
            if self.check_password(pwd, password):
                # return the user Object
                return user
            else:
                raise InvalidAuth('Invalid Credentials')
        except Exception as err:
            raise Exception(err)
        return None

    def set_password(
            self,
            password,
            token_num: int = 6,
            iterations: int = 80000,
            salt: str = None
    ):
        if not salt:
            salt = secrets.token_hex(token_num)
        key = hashlib.pbkdf2_hmac(
            PWD_DIGEST,
            password.encode('utf-8'),
            salt.encode('utf-8'),
            iterations,
            dklen=KEY_LENGTH
        )
        hash = base64.b64encode(key).decode('ascii').strip()
        return f'{PWD_ALGORITHM}${iterations}${salt}${hash}'

    def check_password(self, current_password, password):
        algorithm, iterations, salt, hash = current_password.split('$', 3)
        assert algorithm == PWD_ALGORITHM
        iterations = int(iterations)
        compare_hash = self.set_password(
            password,
            iterations=iterations,
            salt=salt
        )
        return secrets.compare_digest(current_password, compare_hash)


    async def get_payload(self, request):
        ctype = request.content_type
        if request.method == 'GET':
            try:
                user = request.query.get(self.user_attribute, None)
                password = request.query.get(self.pwd_atrribute, None)
                return [user, password]
            except Exception:
                return None
        elif ctype in ('multipart/mixed', 'application/x-www-form-urlencoded'):
            data = await request.post()
            if len(data) > 0:
                user = data.get(self.user_attribute, None)
                password = data.get(self.pwd_atrribute, None)
                return [user, password]
            else:
                return None
        elif ctype == 'application/json':
            try:
                data = await request.json()
                user = data[self.user_attribute]
                password = data[self.pwd_atrribute]
                return [user, password]
            except Exception:
                return None
        else:
            return None

    async def check_credentials(self, request):
        try:
            user, pwd = await self.get_payload(request)
        except Exception as err:
            raise NavException(err, state=400)
        if not pwd and not user:
            raise InvalidAuth('Invalid Credentials', state=401)
        else:
            # making validation
            try:
                user = await self.validate_user(login=user, password=pwd)
            except UserDoesntExists as err:
                raise UserDoesntExists(err)
            except InvalidAuth as err:
                raise InvalidAuth(err, state=401)
            except Exception as err:
                raise NavException(err, state=500)
            try:
                userdata = self.get_userdata(user)
                # Create the User session and returned.
                session = await self._session.create_session(
                    request,
                    user,
                    userdata
                )
                payload = {
                    self.user_property: user[self.userid_attribute],
                    self.username_attribute: user[self.username_attribute],
                    'user_id': user[self.userid_attribute]
                }
                token = self.create_jwt(data=payload)
                return {'token': token}
            except Exception as err:
                print(err)
                return False

    async def authenticate(self, request):
        """ Authenticate, refresh or return the user credentials."""
        pass

    async def auth_middleware(self, app, handler):
        async def middleware(request):
            request.user = None
            authz = await self.authorization_backends(app, handler, request)
            if authz:
                return await authz
            try:
                jwt_token = self.decode_token(request)
            except NavException as err:
                print('Error HERE: ', err, err.state)
                response = {
                    "message": "Token Error",
                    "error": err.message,
                    "status": err.state
                }
                print(response)
                return web.json_response(response, status=err.state)
            except Exception as err:
                raise web.HTTPBadRequest(
                    body=f'Bad Request: {err!s}'
                )
            if self.credentials_required is True:
                raise web.HTTPUnauthorized(
                    body='Unauthorized'
                )
            return await handler(request)
        return middleware



"""Navigator.

Authentication Backends.
"""
from .base import BaseAuthBackend
from .noauth import NoAuth
from .basic import BasicAuth
from .troc import TrocAuth
from .sessionid import SessionIDAuth
from .token import TokenAuth

__all__ = [
    "BaseAuthBackend", "NoAuth",
    "BasicAuth", "TrocAuth",
    "SessionIDAuth", "TokenAuth"
]



import jwt
from aiohttp import web
from datetime import datetime, timedelta
from navigator.conf import (
    SESSION_TIMEOUT,
    SECRET_KEY
)

JWT_SECRET = SECRET_KEY
JWT_ALGORITHM = 'HS256'
JWT_EXP_DELTA_SECONDS = SESSION_TIMEOUT

async def jwt_middleware(app, handler):
    async def middleware(request):
        request.user = None
        jwt_token = request.headers.get('Authorization', None)
        print(jwt_token)
        if jwt_token:
            try:
                payload = jwt.decode(
                    jwt_token,
                    JWT_SECRET,
                    algorithms=[JWT_ALGORITHM]
                )
                print(payload)
            except (jwt.DecodeError, jwt.ExpiredSignatureError) as err:
                print(err)
                return web.json_response(
                    {'message': 'Invalid Token'}, status=400
                )
        return await handler(request)
    return middleware



import logging
import jwt
import importlib
from typing import List, Iterable
from abc import ABC, ABCMeta, abstractmethod
from aiohttp import web, hdrs
from datetime import datetime, timedelta
from asyncdb.utils.models import Model
from navigator.conf import (
    DOMAIN,
    NAV_AUTH_USER,
    NAV_AUTH_GROUP,
    SESSION_NAME,
    SESSION_STORAGE,
    SESSION_TIMEOUT,
    SECRET_KEY,
    JWT_ALGORITHM,
    USER_MAPPING
)
from navigator.auth.session import CookieSession, RedisSession, MemcacheSession
from navigator.exceptions import NavException, UserDoesntExists, InvalidAuth
from navigator.functions import json_response
from aiohttp.web_urldispatcher import SystemRoute

JWT_SECRET = SECRET_KEY
JWT_ALGORITHM = JWT_ALGORITHM
JWT_EXP_DELTA_SECONDS = int(SESSION_TIMEOUT)

exclude_list = (
    '/static/',
    '/api/v1/login',
    '/api/v1/logout',
    '/login',
    '/logout',
    '/signin',
    '/signout',
    '/_debug/'
)


class BaseAuthBackend(ABC):
    """Abstract Base for Authentication."""
    _session = None
    user_model: Model = None
    group_model: Model = None
    user_property: str = 'user'
    user_attribute: str = 'user'
    password_attribute: str = 'password'
    userid_attribute: str = 'user_id'
    username_attribute: str = 'username'
    user_mapping: dict = {'user_id': 'id', 'username': 'username'}
    credentials_required: bool = False
    scheme: str = 'Bearer'
    _authz_backends: List = []
    user_mapping: dict = {}

    def __init__(
            self,
            user_property: str = 'user',
            user_attribute: str = 'user',
            userid_attribute: str = 'user_id',
            username_attribute: str = 'username',
            credentials_required: bool = False,
            authorization_backends: tuple = (),
            session_type: str = 'cookie',
            **kwargs
    ):
        # force using of credentials
        self.credentials_required = credentials_required
        self.user_property = user_property
        self.user_attribute = user_attribute
        self.userid_attribute = userid_attribute
        self.username_attribute = username_attribute
        # authentication scheme
        self._scheme = kwargs['scheme']
        # configuration Authorization Backends:
        self._authz_backends = authorization_backends
        # user and group models
        # getting User and Group Models
        self.user_model = self.get_model(NAV_AUTH_USER)
        self.group_model = self.get_model(NAV_AUTH_GROUP)
        self.user_mapping = USER_MAPPING
        # getting Session Object:
        args = {
            "user_property": user_property,
            "user_attribute": user_attribute,
            "username_attribute": username_attribute,
            **kwargs
        }
        if SESSION_STORAGE == "cookie":
            self._session = CookieSession(name=SESSION_NAME, secret=SECRET_KEY, **args)
        elif SESSION_STORAGE == 'redis':
            self._session = RedisSession(name=SESSION_NAME, **args)
        elif SESSION_STORAGE == 'memcache':
            self._session = MemcacheSession(name=SESSION_NAME, **args)
        else:
            raise Exception(f'Unknown Session type {session_type}')

    def get_model(self, model, **kwargs):
        try:
            parts = model.split('.')
            name = parts[-1]
            classpath = '.'.join(parts[:-1])
            module = importlib.import_module(classpath, package=name)
            obj = getattr(module, name)
            return obj
        except ImportError:
            raise Exception(f"Error loading Auth Model {model}")

    async def get_user(self, **search):
        """Getting User Object."""
        # TODO: getting Groups based on User
        try:
            user = await self.user_model.get(**search)
        except Exception as err:
            logging.error(f'Error getting User {search!s}')
            raise Exception(err)
        # if not exists, return error of missing
        if not user:
            raise UserDoesntExists(f'User doesnt exists')
        return user

    def get_userdata(self, user):
        userdata = {}
        for name, item in self.user_mapping.items():
            if name != self.password_attribute:
                userdata[name] = user[item]
        return userdata

    def configure(self, app, router):
        """ Base configuration for Auth Backends, need to be extended
        to create Session Object."""
        try:
            # configuring Session Object
            self._session.configure_session(app)
        except Exception as err:
            print(err)
            raise Exception(err)

    async def authorization_backends(self, app, handler, request):
        if isinstance(request.match_info.route, SystemRoute):  # eg. 404
            return await handler(request)
        # avoid authorization on exclude list
        if request.path in exclude_list:
            return handler(request)
        # avoid authorization backend on excluded methods:
        if request.method == hdrs.METH_OPTIONS:
            return handler(request)
        # logic for authorization backends
        for backend in self._authz_backends:
            if await backend.check_authorization(request):
                return handler(request)
        return None

    def create_jwt(self, issuer: str = None, expiration: int = None, data: dict = None) -> str:
        """ Creation of JWT tokens based on basic parameters.
        issuer: for default, urn:Navigator
        expiration: in seconds
        **kwargs: data to put in payload
        """
        if not expiration:
            expiration = JWT_EXP_DELTA_SECONDS
        if not issuer:
            issuer = 'urn:Navigator'
        payload = {
            'exp': datetime.utcnow() + timedelta(seconds=expiration),
            "iat": datetime.utcnow(),
            "iss": issuer,
            **data
        }
        jwt_token = jwt.encode(
            payload,
            JWT_SECRET,
            JWT_ALGORITHM,
        )
        return jwt_token

    def decode_token(self, request, issuer: str = None):
        payload = None
        if not issuer:
            issuer = 'urn:Navigator'
        if 'Authorization' in request.headers:
            try:
                scheme, jwt_token = request.headers.get(
                    'Authorization'
                ).strip().split(' ')
            except ValueError as err:
                raise NavException('Invalid authorization Header', state=401)
            if scheme != self.scheme:
                raise NavException('Invalid Session scheme', state=401)
            try:
                payload = jwt.decode(
                    jwt_token,
                    JWT_SECRET,
                    algorithms=[JWT_ALGORITHM],
                    iss=issuer,
                    leeway=30
                )
                logging.debug(f'Decoded Token: {payload!s}')
                return payload
            except (jwt.DecodeError) as err:
                raise NavException(f'Token Decoding Error: {err}', state=400)
            except jwt.InvalidTokenError as err:
                print(err)
                raise NavException(f'Invalid authorization token {err!s}')
            except (jwt.ExpiredSignatureError) as err:
                print(err)
                raise NavException(f'Token Expired {err!s}', state=403)
            except Exception as err:
                print(err, err.__class__.__name__)
                raise NavException(err, state=501)

    async def forgot_session(self, request: web.Request):
        await self._session.forgot_session(request)

    @abstractmethod
    async def check_credentials(self, request):
        """ Authenticate against user credentials (token, user/password)."""
        pass

    @abstractmethod
    async def authenticate(self, request):
        """ Authenticate, refresh or return the user credentials."""
        pass

    async def get_session(self, request):
        """ Get user data from session."""
        app = request.app
        session = await self._session.get_session(request)
        if not session.new:
            # user has existing session
            userdata = session.get(self.user_property)
        else:
            try:
                payload = self.decode_token(request)
            except NavException as err:
                raise NavException(err)
            if payload:
                userdata = payload
        if userdata:
            data = {
                self.user_property: userdata
            }
            return data
        else:
            return None



    @abstractmethod
    async def auth_middleware(self, app, handler):
        """ Base Middleware for Authentication Backend."""
        async def middleware(request):
            return await handler(request)
        return middleware



# -*- coding: utf-8 -*-
import base64
import importlib
import logging
import os
import sys
from pathlib import Path
from types import ModuleType
from typing import Any, Dict, List, Tuple
from cryptography import fernet
# Import Config Class
from navconfig import BASE_DIR, EXTENSION_DIR, config
from navconfig.logging import logdir, loglevel, logging_config

"""
Routes
"""
APP_DIR = BASE_DIR.joinpath("apps")
TEMP_DIR = config.get("TEMP_DIR", fallback="/tmp")
FILES_DIR = config.get("ETL_PATH", fallback="/home/ubuntu/symbits/")
NAV_DIR = BASE_DIR.joinpath("navigator")
STATIC_DIR = BASE_DIR.joinpath("static")
SERVICES_DIR = BASE_DIR.joinpath("services")

"""
Security and debugging
"""
# SECURITY WARNING: keep the secret key used in production secret!
fernet_key = fernet.Fernet.generate_key()
SECRET_KEY = base64.urlsafe_b64decode(fernet_key)
# SECRET_KEY = config.get('TROC_KEY')
PARTNER_KEY = config.get('PARTNER_KEY')
CYPHER_TYPE = config.get('CYPHER_TYPE', fallback='RNC')
HOSTS = [e.strip() for e in list(config.get("HOSTS", fallback="localhost").split(","))]
DOMAIN = config.get('DOMAIN', fallback='dev.local')

# Debug
#
DEBUG = config.getboolean("DEBUG", fallback=True)
PRODUCTION = bool(config.getboolean("PRODUCTION", fallback=(not DEBUG)))
LOCAL_DEVELOPMENT = DEBUG == True and sys.argv[0] == "run.py"
USE_SSL = config.getboolean("ssl", "SSL", fallback=False)

"""
Timezone
"""
# Timezone (For parsedate)
# https://dateparser.readthedocs.io/en/latest/#timezone-and-utc-offset
TIMEZONE = config.get("TIMEZONE", "US/Eastern")

"""
Environment
"""
if DEBUG and LOCAL_DEVELOPMENT:
    ENV = "development"
    CSRF_ENABLED = False
    SSL = False
    SSL_VERIFY = False
    SSL_CERT = None
    SSL_KEY = None
    PREFERRED_URL_SCHEME = "http"
    ENABLE_TOKEN_APP = False
else:
    if PRODUCTION == False and DEBUG == True:
        ENV = "development"
        CSRF_ENABLED = False
    elif PRODUCTION == True and DEBUG == True:
        ENV = "staging"
        CSRF_ENABLED = True
    else:
        ENV = "production"
        CSRF_ENABLED = True
    try:
        SSL_CERT = config.get("CERT")
        SSL_KEY = config.get("KEY")
        PREFERRED_URL_SCHEME = "https"
    except Exception as e:
        SSL_CERT = None
        SSL_KEY = None
        PREFERRED_URL_SCHEME = "http"

# Temp File Path
files_path = BASE_DIR.joinpath("temp")

"""
Basic Information
"""
EMAIL_CONTACT = config.get("EMAIL_CONTACT", section="info", fallback="foo@example.com")
API_NAME = config.get("API_NAME", section="info", fallback="Navigator")

# get settings
from navconfig.conf import *

#######################
##
## APPS CONFIGURATION
##
#######################
"""
Main Database
"""
TIMEZONE = config.get('TIMEZONE', fallback='America/New_York')
PG_USER = config.get('DBUSER')
PG_HOST = config.get('DBHOST', fallback='localhost')
PG_PWD = config.get('DBPWD')
PG_DATABASE = config.get('DBNAME', fallback='navigator')
PG_PORT = config.get('DBPORT', fallback=5432)

asyncpg_url = 'postgres://{user}:{password}@{host}:{port}/{db}'.format(
    user=PG_USER,
    password=PG_PWD,
    host=PG_HOST,
    port=PG_PORT,
    db=PG_DATABASE
)
default_dsn = asyncpg_url

"""
Auth and Cache
"""

"""
REDIS
"""
CACHE_HOST = config.get('CACHEHOST', fallback='localhost')
CACHE_PORT = config.get('CACHEPORT', fallback=6379)
CACHE_URL = "redis://{}:{}".format(CACHE_HOST, CACHE_PORT)
REDIS_SESSION_DB = config.get('REDIS_SESSION_DB', fallback=0)

"""
Authentication System
"""
NAV_AUTH_BACKEND = config.get('AUTH_BACKEND', fallback='navigator.auth.backends.NoAuth')
AUTHORIZATION_BACKENDS = [e.strip() for e in list(config.get("AUTHORIZATION_BACKENDS", fallback="allow_hosts").split(","))]
CREDENTIALS_REQUIRED = config.getboolean('AUTH_CREDENTIALS_REQUIRED', fallback=False)
NAV_AUTH_USER = config.get('AUTH_USER_MODEL', fallback='navigator.auth.models.User')
NAV_AUTH_GROUP = config.get('AUTH_GROUP_MODEL', fallback='navigator.auth.models.Group')
USER_MAPPING = {
    "user_id": "user_id",
    "username": "username",
    "password": "password",
    "first_name": "first_name",
    "last_name": "last_name",
    "email": "email",
    "enabled": "is_active",
    "superuser": "is_superuser",
    "last_login": "last_login",
    "title": "title"
}
USERS_TABLE = config.get('AUTH_USERS_TABLE', fallback='vw_users')
ALLOWED_HOSTS = [
    e.strip() for e in list(config.get("ALLOWED_HOSTS", fallback="localhost*").split(","))
]
"""
Session Storage
"""
SESSION_STORAGE = config.get('SESSION_STORAGE', fallback='redis')
SESSION_URL = "redis://{}:{}/{}".format(CACHE_HOST, CACHE_PORT, REDIS_SESSION_DB)
CACHE_PREFIX = config.get('CACHE_PREFIX', fallback='navigator')
SESSION_PREFIX = '{}_session'.format(CACHE_PREFIX)
SESSION_NAME = '{}_SESSION'.format(config.get('APP_TITLE', fallback='NAVIGATOR').upper())
SESSION_TIMEOUT = config.get('SESSION_TIMEOUT', fallback=3600)
JWT_ALGORITHM = config.get('JWT_ALGORITHM', fallback='HS256')

"""
 Memcache
"""
MEMCACHE_HOST = config.get('MEMCACHE_HOST', 'localhost')
MEMCACHE_PORT = config.get('MEMCACHE_PORT', 11211)

"""
Final: Config dict for aiohttp
"""
Context = {
    "DEBUG": DEBUG,
    "DEVELOPMENT": (not PRODUCTION),
    "LOCAL_DEVELOPMENT": LOCAL_DEVELOPMENT,
    "PRODUCTION": PRODUCTION,
    "SECRET_KEY": SECRET_KEY,
    "env": ENV,
    "cache_url": CACHE_URL,
    "asyncpg_url": asyncpg_url,
    "default_dsn": default_dsn
}

"""
Applications
"""
INSTALLED_APPS: List = []
DATABASES: Dict = {}

if APP_DIR.is_dir():
    for item in APP_DIR.iterdir():
        # for item in os.listdir(APPS_DIR):
        if item.name != "__pycache__":
            if item.is_dir():
                name = item.name
                if not name in INSTALLED_APPS:
                    app_name = "apps.{}".format(item.name)
                    path = APP_DIR.joinpath(name)
                    url_file = path.joinpath("urls.py")
                    try:
                        i = importlib.import_module(app_name, package="apps")
                        if isinstance(i, ModuleType):
                            # is a Navigator Program
                            INSTALLED_APPS += (app_name,)
                    except ImportError as err:
                        print("ERROR: ", err)
                        continue
                    # schema configuration
                    DATABASES[item.name] = {
                        #"ENGINE": config.get("DBENGINE"),
                        "NAME": PG_DATABASE,
                        "USER": PG_USER,
                        "OPTIONS": {
                            "options": "-c search_path=" + item.name + ",public",
                        },
                        #'PARAMS': {
                        #    'readonly': True,
                        # },
                        "SCHEMA": item.name,
                        "PASSWORD": PG_PWD,
                        "HOST": PG_HOST,
                        "PORT": PG_PORT,
                    }

Context['DATABASES'] = DATABASES

"""
Per-Program Settings
"""
program: str
for program in INSTALLED_APPS:
    settings = "apps.{}.settings".format(program)
    try:
        i = importlib.import_module(settings)
        globals()[program] = i
    except ImportError as err:
        pass



""" Abstract Class for create Session Objects."""
import time
import asyncio
import logging
import base64
from functools import wraps, partial
from cryptography import fernet
from abc import ABC, abstractmethod
# aiohttp session
from aiohttp_session import get_session, new_session
from navigator.auth.models import User
from navigator.conf import (
    DOMAIN,
    SESSION_URL,
    SESSION_NAME,
    SESSION_PREFIX,
    SESSION_TIMEOUT
)

class AbstractSession(ABC):
    """Abstract Base Session."""
    session = None
    session_name: str = 'NAVIGATOR_SESSION'
    secret_key: str = None
    user_property: str = 'user'
    user_attribute: str = 'user_id'
    username_attribute: str = 'username'

    def __init__(
            self,
            secret: str = '',
            name: str = '',
            user_property: str = 'user',
            user_attribute: str = 'user_id',
            username_attribute: str = 'username',
            **kwargs
    ):
        if name:
            self.session_name = name
        else:
            self.session_name = SESSION_NAME
        if not secret:
            fernet_key = fernet.Fernet.generate_key()
            self.secret_key = base64.urlsafe_b64decode(fernet_key)
        else:
            self.secret_key = secret
        # user property:
        self.user_property = user_property
        self.user_attribute = user_attribute
        self.username_attribute = username_attribute

    @abstractmethod
    async def configure_session(self, app):
        pass

    async def get_session(self, request):
        return await get_session(request)

    async def create_session(self, request, user, userdata):
        app = request.app
        try:
            session = await new_session(request)
        except Exception as err:
            print(err)
            return False
        last_visit = session["last_visit"] if "last_visit" in session else None
        session["last_visit"] = time.time()
        session["last_visited"] = "Last visited: {}".format(last_visit)
        # think about saving user data on session when create
        app['User'] = user
        app[self.user_property] = userdata
        session[self.user_property] = userdata
        app["session"] = self.session
        # logging.debug('Creating Session: {}'.format(session))
        return session

    async def forgot_session(self, request):
        app = request.app
        session = await get_session(request)
        session.invalidate()
        try:
            app['User'] = None
            app[self.user_property] = None
            request.user = None
        except Exception as err:
            print(err)
        app["session"] = None



"""Session based on REDIS Server."""
import asyncio
import json
import rapidjson
# redis pool
import aioredis
from functools import wraps, partial
from aiohttp_session.redis_storage import RedisStorage
from aiohttp_session import setup as setup_session
from .base import AbstractSession
from navigator.conf import (
    DOMAIN,
    SESSION_URL,
    SESSION_TIMEOUT
)


class RedisSession(AbstractSession):
    """Session Storage based on Redis."""
    _pool = None

    async def get_redis(self, **kwargs):
        kwargs['timeout'] = 1
        loop = asyncio.get_event_loop()
        return await aioredis.create_redis_pool(SESSION_URL, loop=loop, **kwargs)

    def configure_session(self, app, **kwargs):
        async def _make_mredis():
            _encoder = partial(rapidjson.dumps, datetime_mode=rapidjson.DM_ISO8601)
            try:
                self._pool = await self.get_redis()
                setup_session(
                    app,
                    RedisStorage(
                        self._pool,
                        cookie_name=self.session_name,
                        encoder=_encoder,
                        decoder=rapidjson.loads,
                        max_age=int(SESSION_TIMEOUT)
                    )
                )
            except Exception as err:
                print(err)
                return False
        return asyncio.get_event_loop().run_until_complete(
            _make_mredis()
        )
